/* A simple server in the internet domain using TCP
   The port number is passed as an argument 
   This version runs forever, forking off a separate 
   process for each connection
   gcc server2.c -lsocket
*/
#include <stdio.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include<stdlib.h>
#include<unistd.h>
#include<strings.h>
#include<string.h>
#include<pthread.h>	
#include "toxml.h"
#include "KVCache.c"

/*JOB QUEUE POINTER*/
jobs_queue *job_head=NULL;
jobs_queue *job_tail=NULL;

/*CACHE POINTERS*/
kvcache_data **t;
int *repl;

/*LOCK FOR INSERTING IN JOB QUEUE*/
pthread_mutex_t job_q_lock;

/* FUNCTION PROTOTYPE OF PROCESSING QUERY*/
void process_work(jobs_queue*);

/* FUNCION PROTOTYPE FOR SERVING THE CLIENT */
void *serve(void*); 

/* FUNCTION FOR HANDLING ERRORS*/
void error(char *msg)
{
    perror(msg);
    exit(1);
}

int main(int argc, char *argv[])
{
     int sockfd, newsockfd, portno, clilen, pid;
     struct sockaddr_in serv_addr, cli_addr;
     t= init_cache();
     repl=rep_line_per_set();
	 pthread_mutex_init(&job_q_lock,NULL);
     sockfd = socket(AF_INET, SOCK_STREAM, 0);
     if (sockfd < 0) 
        error("ERROR opening socket");
     bzero((char *) &serv_addr, sizeof(serv_addr));
     serv_addr.sin_family = AF_INET;
     serv_addr.sin_addr.s_addr = INADDR_ANY;
     serv_addr.sin_port = htons(8080);
     if (bind(sockfd, (struct sockaddr *) &serv_addr,
              sizeof(serv_addr)) < 0) 
              error("ERROR on binding");
     listen(sockfd,5);
     clilen = sizeof(cli_addr);
     while (1) {
         newsockfd = accept(sockfd, 
               (struct sockaddr *) &cli_addr, &clilen);
      
         pthread_t thr;
         int *arg;
         arg=malloc(sizeof(*arg));
         *arg=newsockfd;
         pthread_create(&thr,NULL,serve,arg);
     } /* end of while */
     return 0; /* we never get here */
}

/******** serve() *********************
 There is a separate instance of this function 
 for each connection.  It handles all communication
 once a connnection has been established.
 *****************************************/
void *	serve (void* socket)
{

   int sock;
   int n;
   long tid=pthread_self();
   sock=*((int*) socket);
   char buffer[300000];
     
   while(1){
   //zeroing all the value in buffer
   bzero(buffer,300000);
   
   //reading the socket
   n = read(sock,buffer,300000);
   if (n < 0) error("ERROR reading from socket");
   else if (n==0) {
   	printf("Client has been closed,Killing the thread\n");
   	pthread_cancel(tid);
   }
   else{
   //Creating a structure to add in linked list of job queue
   jobs_queue *job=(jobs_queue *)malloc((sizeof(jobs_queue)));
   job->fd=sock;
   job->next=NULL;
 
   fromXML(job->key,job->value,job->messagetype,job->message,buffer);
   
   //Adding to the chain in Job Queue
   pthread_mutex_lock(&job_q_lock);
   if(job_head==NULL){
   		job_head=job;
   		job_tail=job;
   }
   else{
   		job_tail->next=job;
   		job_tail=job;
   }
   pthread_mutex_unlock(&job_q_lock);
   process_work(job);
   }
   }
}


/**********process_work******************
Function to process query in job_queue
and then writing to the client
****************************************/
void process_work(jobs_queue *job){
	//printf("executing job\n");
	char resp[300000];
	if(!strcmp(job->messagetype,"putreq")){
		if(postInCache(t,repl,job->key,job->value)){
			toXML(resp,"","","resp","Success");
		}
		else{
				toXML(resp,"","","resp","IOError");
		
		}
	}
	else if(!strcmp(job->messagetype,"getreq")){
		if(getInCache(t,repl,job->key,job->value)){
				toXML(resp,job->key,job->value,"resp","");	
			}
		else{
			toXML(resp,"","","resp","Does not exist");
		}
	}
	else if(!strcmp(job->messagetype,"delreq")){
		if(delInCache(t,repl,job->key)){
			toXML(resp,"","","resp","Success");
		}
		else{
			toXML(resp,"","","resp","Does not exist");
		
		}
	}
	//printf("%s\n",resp);
	write(job->fd,resp,strlen(resp));

}
